package edu.ucla.cs.cs144;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.PreparedStatement;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.Hit;
import org.apache.lucene.search.Hits;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;

import java.util.Date;
import java.util.Iterator;
import java.util.*;
import java.io.*;
import java.text.SimpleDateFormat;
import java.text.DecimalFormat;
import java.util.Locale;

import edu.ucla.cs.cs144.DbManager;
import edu.ucla.cs.cs144.SearchConstraint;
import edu.ucla.cs.cs144.SearchResult;
import edu.ucla.cs.cs144.SearchEngine;

public class AuctionSearch implements IAuctionSearch 
{	
	/* Function to perform basic search over the Lucene indexes on the union of name-description-categories */
	public SearchResult[] basicSearch(String query, int numResultsToSkip, int numResultsToReturn)
	{
		Vector <SearchResult> resultsVector = new Vector <SearchResult>();	//Vector to collect items that hit.
		SearchResult[] searchResults = null;
		
		//Verify only positive values in the entries.
		numResultsToSkip = (numResultsToSkip < 0)? 0: numResultsToSkip;
		numResultsToReturn = (numResultsToReturn < 0)? 0: numResultsToReturn;
		
		try
		{
			Hits hits = SearchEngine.performSearch(query);	//Search over the content field.
			
			for(int I=0; I<hits.length(); I++)				//Iterate over hit documents.
			{
				Document doc = hits.doc(I);					//Get documents in hits.
				SearchResult sr = new SearchResult(doc.get("itemID"), "");	//No name for now.
				resultsVector.add(sr);
			}
		}
		catch(IOException e)
		{
			System.err.println("Failed to read the inverted index files: "+e.getMessage());
		}
		catch(ParseException p)
		{
			//System.err.println("Failed to perform searching on the inverted index: "+p.getMessage());
			return new SearchResult[0];		//In case of mal-formed keyword, return 0 results.
		}
		
		//Trim to the values specified in numResultsToSkip and numResultsToReturn.
		if(numResultsToSkip > resultsVector.size())
			numResultsToSkip = resultsVector.size();	//Check for right quantities to remove.
		
		for(int I=0; I<numResultsToSkip; I++)			//Skip the determined number of hits.
			resultsVector.remove(0);					
		
		if(resultsVector.size() < numResultsToReturn || numResultsToReturn == 0)
			numResultsToReturn = resultsVector.size();	//Check if there are enough records to show.
		
		searchResults = new SearchResult[numResultsToReturn];
		
		if(resultsVector.size() > 0)					//Are there still elements?
		{			
			//////////////// Prepare to load missing name from the item relation ///////////////
			try 
	        {
	        	Connection conn = DbManager.getConnection(true);	//Connect to database.
	        	PreparedStatement pst = conn.prepareStatement("SELECT name FROM item WHERE itemID=?;");
	        	
	        	for(int I=0; I<searchResults.length; I++)	//Load name for every itemID.
	        	{
	        		searchResults[I] = resultsVector.remove(0);		//Starts from the beginning of the vector.
	        		pst.setString(1, searchResults[I].getItemId());	//Set parameter for new search.
	        		ResultSet rs = pst.executeQuery();
	        		if(rs.next())							//Read the name.
	        			searchResults[I].setName(rs.getString("name"));
	        		else
	        			System.err.println("Could not find the name for item "+searchResults[I].getItemId());
	        	}
	        	
	        	conn.close();								//Close connection to database.
	        }
			catch(SQLException e)
			{
				System.err.println("Failed to load names for items: "+e.getMessage());
			}
		}
				
		return searchResults;	//It always return an array of whatever size, even zero.
	}

	/* Function to perform advance searches */
	public SearchResult[] advancedSearch(SearchConstraint[] constraints, int numResultsToSkip, int numResultsToReturn) 
	{
		SearchResult[] searchResults = null;
		Vector <SearchResult> resultsVector = new Vector <SearchResult>();
		Set <String> resultSet = new LinkedHashSet <String>();	//Initialize it to make sure its size is 0.
		Set <String> nameSet = new LinkedHashSet <String>();	//Sets to store itemIDs from the searchs.
		Set <String> categorySet = new LinkedHashSet <String>();
		Set <String> descriptionSet = new LinkedHashSet <String>();

		Vector <String> sellerIDVector = new Vector <String>();	//Vectors to collect multiple constraints.
		Vector <String> bidderIDVector = new Vector <String>();
		Vector <String> buyPriceVector = new Vector <String>();
		Vector <String> endsVector = new Vector <String>();
		Vector <String> nameVector = new Vector <String>();
		Vector <String> categoryVector = new Vector <String>();
		Vector <String> descriptionVector = new Vector <String>();
		
		Set <String> sqlSet = new LinkedHashSet <String>();		//Set for item ids from sql indices.
		Connection conn = null;									//Connection to database.
		
		//Verify only positive values in the entries.
		numResultsToSkip = (numResultsToSkip < 0)? 0: numResultsToSkip;
		numResultsToReturn = (numResultsToReturn < 0)? 0: numResultsToReturn;
		
		/////////////////// Detect what constraints are given //////////////
		for(int I=0; I<constraints.length; I++)
		{
			if(constraints[I].getValue().trim().length() > 0)	//Check that something is given.
			{
				if(constraints[I].getFieldName().equals(FieldName.ItemName))	//Item name?
					nameVector.add(constraints[I].getValue());
				if(constraints[I].getFieldName().equals(FieldName.Category))	//Category?
					categoryVector.add(constraints[I].getValue());
				if(constraints[I].getFieldName().equals(FieldName.Description))	//Item description?
					descriptionVector.add(constraints[I].getValue());
				if(constraints[I].getFieldName().equals(FieldName.SellerId))	//Item seller id?
					sellerIDVector.add(constraints[I].getValue().trim());
				if(constraints[I].getFieldName().equals(FieldName.BidderId))	//Auction bidder?
					bidderIDVector.add(constraints[I].getValue().trim());
				if(constraints[I].getFieldName().equals(FieldName.BuyPrice))	//Item buy price?
					buyPriceVector.add(constraints[I].getValue().trim());
				if(constraints[I].getFieldName().equals(FieldName.EndTime))		//Auction end time?
					endsVector.add(auctionDateToTimestamp(constraints[I].getValue().trim()));
			}
		}
		
		///////////// Compute the sets for each of the textual constraints ////////////
		if(nameVector.size()>0)			//Were there names?
			nameSet = searchText("name", stringifyVector("", nameVector));
		if(categoryVector.size()>0)		//Were there categories.
			categorySet = searchText("categories", stringifyVector("", categoryVector));
		if(descriptionVector.size()>0)	//Were there descriptions.
			descriptionSet = searchText("description", stringifyVector("", descriptionVector));
		
		///////////// Compute the set of item ids from the SQL indices /////////////
		if(sellerIDVector.size()>0 || endsVector.size()>0 || buyPriceVector.size()>0 || bidderIDVector.size()>0)
		{
			String select = "SELECT ";
			String from = "FROM ";
			String where = "WHERE ";
			int tables = 0;					//Keeps the number of tables involved in the query.
			int conditions = 0;
			String lastTable = "";
			
			if(sellerIDVector.size()>0 || endsVector.size()>0)	//item table required?
			{
				from += "item"; tables++; lastTable="item";
				if(sellerIDVector.size()>0)
				{
					//where += "sellerID='" + sellerID + "'";
					where += stringifyVector("sellerID", sellerIDVector);
					conditions++;
				}
				if(endsVector.size()>0)
				{
					//where += (conditions>0)? " AND ends='" + ends + "'": "ends='" + ends + "'";
					where += (conditions>0)? " AND " + stringifyVector("ends", endsVector): stringifyVector("ends", endsVector);
					conditions++;
				}
			}
			if(buyPriceVector.size()>0)					//buy price table required?
			{
				from += (tables>0)? ", itemBuyPrice": "itemBuyPrice";
				where += (tables>0)? " AND itemBuyPrice.itemID=" + lastTable + ".itemID": "";
				tables++; lastTable="itemBuyPrice";
				//where += (conditions>0)? " AND buyPrice=" + buyPrice: "buyPrice=" + buyPrice;
				where += (conditions>0)? " AND " + stringifyVector("buyPrice", buyPriceVector): stringifyVector("buyPrice", buyPriceVector);
				conditions++;
			}
			if(bidderIDVector.size()>0)					//bidInfo table required?
			{
				from += (tables>0)? ", bidInfo": "bidInfo";
				where += (tables>0)? " AND bidInfo.itemID=" + lastTable + ".itemID": "";
				tables++; lastTable="bidInfo";
				//where += (conditions>0)? " AND bidderID='" + bidderID + "'": "bidderID='" + bidderID + "'";
				where += (conditions>0)? " AND " + stringifyVector("bidderID", bidderIDVector): stringifyVector("bidderID", bidderIDVector);
				conditions++;
			}
			
			where += ";";							//Completes sql query.
			select += lastTable+".itemID AS itemID";
			
			try
			{
				conn = DbManager.getConnection(true);	//Connect to database.
	        	PreparedStatement pst = conn.prepareStatement(select + " " + from + " " + where);
	        	//System.out.println("Indices' SQL query: "+select + " " + from + " " + where);
	        	ResultSet rs = pst.executeQuery();
	    		while(rs.next())									//Read the item ids.
	    		{
	    			sqlSet.add(rs.getString("itemID"));				//Adds item ids to the sql set.
	    		}
			}
			catch(SQLException e)
			{
				System.err.println("Failed to perform query on indexed fields in relations: "+e.getMessage());
			}
		}
		
		//////////////// Compute intersection set with item ids from all given fields /////////////////
		if(nameVector.size()>0)	//Case when name is given.
		{
			resultSet = new LinkedHashSet <String>(nameSet);
			if(sellerIDVector.size()>0 || bidderIDVector.size()>0 || endsVector.size()>0 || buyPriceVector.size()>0) resultSet.retainAll(sqlSet);
			if(categoryVector.size()>0) resultSet.retainAll(categorySet);
			if(descriptionVector.size()>0) resultSet.retainAll(descriptionSet);
		}
		else
		{
			if(sellerIDVector.size()>0 || bidderIDVector.size()>0 || endsVector.size()>0 || buyPriceVector.size()>0)	//Case when a sql index is given.
			{
				resultSet = new LinkedHashSet <String>(sqlSet);
				if(nameVector.size()>0) resultSet.retainAll(nameSet);
				if(categoryVector.size()>0) resultSet.retainAll(categorySet);
				if(descriptionVector.size()>0) resultSet.retainAll(descriptionSet);
			}
			else
			{
				if(categoryVector.size()>0)	//Case when category is given.
				{
					resultSet = new LinkedHashSet <String>(categorySet);
					if(sellerIDVector.size()>0 || bidderIDVector.size()>0 || endsVector.size()>0 || buyPriceVector.size()>0) resultSet.retainAll(sqlSet);
					if(nameVector.size()>0) resultSet.retainAll(nameSet);
					if(descriptionVector.size()>0) resultSet.retainAll(descriptionSet);
				}
				else
				{
					if(descriptionVector.size()>0)	//Case when description is given.
					{
						resultSet = new LinkedHashSet <String>(descriptionSet);
						if(sellerIDVector.size()>0 || bidderIDVector.size()>0 || endsVector.size()>0 || buyPriceVector.size()>0) resultSet.retainAll(sqlSet);
						if(nameVector.size()>0) resultSet.retainAll(nameSet);
						if(categoryVector.size()>0) resultSet.retainAll(categorySet);
					}
				}
			}
		}
		
		//////////////// Pull names and item ids from the data base into searchResults /////////////
		Iterator <String> setIterator = resultSet.iterator();
		int skipCount = 0, returnCount = 0;
		
		if(numResultsToReturn == 0)		//Return all results after skipping some of them?
		{
			numResultsToReturn = resultSet.size() - numResultsToSkip;
		}
		
		try
		{
			if(conn == null)	//There were not sql indices involved.
				conn = DbManager.getConnection(true);	//Connect to database.
			PreparedStatement pst = conn.prepareStatement("SELECT name FROM item WHERE itemID=?");
			
			while(setIterator.hasNext() && returnCount < numResultsToReturn) //Go over all string elements in the set.
			{
				String itemID = setIterator.next();
				if(skipCount < numResultsToSkip)		//Skip first results.
					skipCount++;
				else
				{
					pst.setString(1, itemID);		//Set parameter for new search.
	        		ResultSet rs = pst.executeQuery();
	        		if(rs.next())					//Read the name.
	        		{
	        			SearchResult sr = new SearchResult(itemID, rs.getString("name"));
	        			resultsVector.add(sr);		//Fill in a dynamic vector.
	        		}
	        		else
	        			System.err.println("Could not find the name for item "+itemID);
	        		returnCount++;
				}
			}
        	
			searchResults = new SearchResult[resultsVector.size()];	//Allocate memory for the resulting array.
			resultsVector.copyInto(searchResults);
			
			conn.close();											//Close connection to data base.
		}
		catch(SQLException e)
		{
			System.err.println("Failed to get the names for the item ids: "+e.getMessage());
		}
		
		return searchResults;	//It always return an array of whatever size, even zero.
	}

	/* Function to create an XML string of a given itemID */
	public String getXMLDataForItemId(String itemId) 
	{
		StringBuffer buffer = new StringBuffer("");
		Connection conn = null;						//Variables to connect to database and get info.
		PreparedStatement itemSt = null;
		PreparedStatement itemBuyPriceSt = null;
		PreparedStatement itemCategorySt = null;
		PreparedStatement userRatingSt = null;
		PreparedStatement userLocSt = null;
		PreparedStatement userCountrySt = null;
		PreparedStatement bidInfoSt = null;
		
		Locale.setDefault(Locale.US);				//Format for quantities.
        DecimalFormat df = new DecimalFormat("0.00");
		
		try
		{
			conn = DbManager.getConnection(true);	//Connect to CS144 database.
			
			itemSt = conn.prepareStatement("SELECT * FROM item WHERE itemID=?;");
			itemSt.setString(1, itemId);
			ResultSet itemRS = itemSt.executeQuery();
			
			if(itemRS.next())						//Requested ItemId exists?
			{
				//buffer.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>" + "\n");	//Writes first line of the XML string.
				
				buffer.append("<Item ItemID=\"" + itemRS.getString("itemID") + "\" >" + "\n");	//Writes Item header.
				
				buffer.append("\t<Name>" + parseToXML(itemRS.getString("name")) + "</Name>" + "\n");		//Item name.

				itemCategorySt = conn.prepareStatement("SELECT category FROM itemCategory WHERE itemID=?;");
				itemCategorySt.setString(1, itemRS.getString("itemID"));
				ResultSet categoryRS = itemCategorySt.executeQuery();
				while(categoryRS.next())								//Read all categories.
				{
					buffer.append("\t<Category>" + parseToXML(categoryRS.getString("category")) + "</Category>" + "\n");
				}
				
				buffer.append("\t<Currently>$" + df.format(itemRS.getDouble("currently")) + "</Currently>" + "\n");	//Currently value.
				
				itemBuyPriceSt = conn.prepareStatement("SELECT buyPrice FROM itemBuyPrice WHERE itemID=?;");
				itemBuyPriceSt.setString(1, itemRS.getString("itemID"));
				ResultSet buyPriceRS = itemBuyPriceSt.executeQuery();
				if(buyPriceRS.next())		//May or may not be in the item data.
				{
					buffer.append("\t<Buy_Price>$" + df.format(buyPriceRS.getDouble("buyPrice")) + "</Buy_Price>" + "\n");	//Buy_Price.
				}
				
				buffer.append("\t<First_Bid>$" + df.format(itemRS.getDouble("firstBid")) + "</First_Bid>" + "\n");	//First_Bid value.
				
				buffer.append("\t<Number_of_Bids>" + itemRS.getInt("numberOfBids") + "</Number_of_Bids>" + "\n");	//Number_of_Bids value.
				
				userRatingSt = conn.prepareStatement("SELECT * FROM userRating WHERE userID=?;");		//Prepare statements for using them in
				userLocSt = conn.prepareStatement("SELECT * FROM userLoc WHERE userID=?;");				//bidder and seller (after bids) data
				userCountrySt = conn.prepareStatement("SELECT * FROM userCountry WHERE userID=?;");		//extraction.
				
				if(itemRS.getInt("numberOfBids") > 0)		//Are there bids?
				{
					buffer.append("\t<Bids>" + "\n");			//Bids root.
					
					bidInfoSt = conn.prepareStatement("SELECT * FROM bidInfo WHERE bidInfo.itemID=? ORDER BY time ASC;");
					bidInfoSt.setString(1, itemRS.getString("itemID"));
					ResultSet bidInfoRS = bidInfoSt.executeQuery();
					
					while(bidInfoRS.next())					//Find all bids for this item.
					{
						buffer.append("\t\t<Bid>" + "\n");	//Bid root.
						
						userRatingSt.setString(1, parseToXML(bidInfoRS.getString("bidderID")));		//Bidder data.
						ResultSet userRatingRS = userRatingSt.executeQuery();
						userRatingRS.next();
						buffer.append("\t\t\t<Bidder UserID=\"" + parseToXML(bidInfoRS.getString("bidderID")) + "\" Rating=\"" + userRatingRS.getInt("rating") + "\" ");
						
						boolean hasLocation = false, hasCountry = false;		//Variables to know when to leave <bidder />.
						String toAppendLocation = "", toAppendCountry = "";
						
						userLocSt.setString(1, parseToXML(bidInfoRS.getString("bidderID")));
						ResultSet userLocRS = userLocSt.executeQuery();
						if(userLocRS.next())				//Bidder has location?
						{
							toAppendLocation = "\t\t\t\t<Location>" + parseToXML(userLocRS.getString("location")) + "</Location>" + "\n";
							hasLocation = true;
						}
						
						userCountrySt.setString(1, parseToXML(bidInfoRS.getString("bidderID")));
						ResultSet userCountryRS = userCountrySt.executeQuery();
						if(userCountryRS.next())			//Bidder has country?
						{
							toAppendCountry = "\t\t\t\t<Country>" + parseToXML(userCountryRS.getString("Country")) + "</Country>" + "\n";
							hasCountry = true;
						}
						
						//Determine if there were either location or country.
						if(hasCountry || hasLocation)
						{
							buffer.append(">" + "\n");					//Close correctly the opening bidder tag.
							buffer.append(toAppendLocation);			//Add both location and country.
							buffer.append(toAppendCountry);				//If one is not given, nothing is attached.
							buffer.append("\t\t\t</Bidder>" + "\n");	//Closing tag for bidder.
						}
						else
							buffer.append("/>" + "\n");					//Close correctly the opening and closing bidder tag.				
						
						buffer.append("\t\t\t<Time>" + timestampToAuctionDate(bidInfoRS.getString("time")) + "</Time>" + "\n");	//Bid time.
						
						buffer.append("\t\t\t<Amount>$" + df.format(bidInfoRS.getDouble("amount")) + "</Amount>" + "\n");			//Get bid amount.
						
						buffer.append("\t\t</Bid>" + "\n");	//Closing tag for bid root.
					}
					
					buffer.append("\t</Bids>" + "\n");		//Bids root closing tag.
				}
				
				userLocSt.setString(1, parseToXML(itemRS.getString("sellerID")));
				ResultSet userLocRS = userLocSt.executeQuery();
				userLocRS.next();
				buffer.append("\t<Location>" + parseToXML(userLocRS.getString("location")) + "</Location>" + "\n");	//Auction location.
				
				userCountrySt.setString(1, parseToXML(itemRS.getString("sellerID")));
				ResultSet userCountryRS = userCountrySt.executeQuery();
				userCountryRS.next();
				buffer.append("\t<Country>" + parseToXML(userCountryRS.getString("Country")) + "</Country>" + "\n");	//Auction country.
				
				buffer.append("\t<Started>" + timestampToAuctionDate(itemRS.getString("started")) + "</Started>" + "\n");	//Auction starting time.
				
				buffer.append("\t<Ends>" + timestampToAuctionDate(itemRS.getString("ends")) + "</Ends>" + "\n");	//Auction ending time.
				
				userRatingSt.setString(1, parseToXML(itemRS.getString("sellerID")));		//Seller data.
				ResultSet userRatingRS = userRatingSt.executeQuery();
				userRatingRS.next();
				buffer.append("\t<Seller UserID=\"" + parseToXML(itemRS.getString("sellerID")) + "\" Rating=\"" + userRatingRS.getInt("rating") + "\" />" + "\n");
				
				if(itemRS.getString("description").length()==0)	//No description?
					buffer.append("\t<Description />" + "\n");
				else
					buffer.append("\t<Description>" + parseToXML(itemRS.getString("description")) + "</Description>" + "\n");	//Item description.
				
				buffer.append("</Item>");															//Close Item tag.
			}
		}
		catch(SQLException e)
		{
			System.err.println("Failed to fetch data for XML composing: " + e.getMessage());
		}
		
		return buffer.toString();
	}
	
	public String echo(String message) 
	{
		return message;
	}
	
	/* Function to stringify a vector of strings plus an equality condition */
	private String stringifyVector(String eq, Vector <String> vector)
	{
		StringBuffer sb = new StringBuffer("");
		if(vector.size()>0)		//Check the vector has elements.
		{
			for(int I=0; I<vector.size(); I++)	//Collect each element.
			{
				if(I > 0)
					sb.append(" AND "); 		//Compose a string.
				
				if(eq.length() > 0)				//A equality expression given?
					sb.append(eq + "='" + vector.elementAt(I) + "'");
				else
					sb.append(vector.elementAt(I));	//Only add the word
			}
		}
		return sb.toString();	//Return a valid concatenation if vector has elements, empty string otherwise.
	}
	
	/* Function to search keywords in text fields using Lucene indices */
	private Set <String> searchText(String field, String keyword)
	{
		Set <String> resultsSet = new LinkedHashSet <String>();
		String query = field + ":( " + keyword + " )";
		
		try
		{
			Hits hits = SearchEngine.performSearch(query);	//Search over the given field.
			
			for(int I=0; I<hits.length(); I++)				//Iterate over hit documents.
			{
				Document doc = hits.doc(I);					//Get documents in hits.
				resultsSet.add(doc.get("itemID"));
			}
		}
		catch(IOException e)
		{
			System.err.println("Failed to read the inverted index files: "+e.getMessage());
		}
		catch(ParseException p)
		{
			//System.err.println("Failed to perform searching on the inverted index: "+p.getMessage());
			return resultsSet;		//Return an empty set.
		}
		
		return resultsSet;
	}
	
	/* Function to parse the auction date format into TIMESTAMP MySQL format */
    private String auctionDateToTimestamp(String dateStr)
    {
    	SimpleDateFormat auctionDateFormat = new SimpleDateFormat("MMM-dd-yy HH:mm:ss");
    	SimpleDateFormat timestampFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
    	
    	Date auctionDate = new Date();
    	try
    	{
    		auctionDate = auctionDateFormat.parse(dateStr);		//Create a Date object.
    	}
    	catch(java.text.ParseException e)
    	{
    		System.err.println("Failed to parse auction date.");
    	}
    	return timestampFormat.format(auctionDate);					//Return the equivalence in timestamp.
    }
    
    /* Function to parse the TIMESTAMP MySQL format into auction date format */
    private String timestampToAuctionDate(String dateStr)
    {
    	SimpleDateFormat auctionDateFormat = new SimpleDateFormat("MMM-dd-yy HH:mm:ss");
    	SimpleDateFormat timestampFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
    	
    	Date sqlDate = new Date();
    	try
    	{
    		sqlDate = timestampFormat.parse(dateStr);		//Create a Date object.
    	}
    	catch(java.text.ParseException e)
    	{
    		System.err.println("Failed to parse timestamp sql date.");
    	}
    	return auctionDateFormat.format(sqlDate);					//Return the equivalence in timestamp.
    }
    
    /* Function to parse a common string to a valid XML string */
    private String parseToXML(String str)
    {
    	StringBuffer result = new StringBuffer();
    	
    	for(int I=0; I<str.length(); I++)		//Iterate over each character of the string.
    	{
    		switch(str.charAt(I))				//Compare character with XML spetial character.
    		{
    		case '\"': result.append("&quot;"); break;
    		case '\'': result.append("&apos;"); break;
    		case '&': result.append("&amp;"); break;
    		case '<': result.append("&lt;"); break;
    		case '>': result.append("&gt;"); break;
    		default: result.append(str.charAt(I));
    		}
    	}
    	
    	return result.toString();
    }
}
